
local lib = require "lib"

local coin_tiers = {}



-- Initialize a new coin object
function coin_tiers.new()
    -- Create the coin object with configuration
    local coin = {
        tier_scaling = 100000,
        max_coin_tier = 4,
        values = {}
    }
    
    -- Initialize values (default to zero)
    for i = 1, coin.max_coin_tier do
        coin.values[i] = 0
    end
    
    return coin
end

-- Normalize a coin, ensuring all tiers are within bounds and have integer values
function coin_tiers.normalize(coin)
    local tier_scaling = coin.tier_scaling

    -- Convert decimals to integers
    for i = coin.max_coin_tier, 1, -1 do
        if coin.values[i] ~= math.floor(coin.values[i]) then
            if i > 1 then
                coin.values[i - 1] = coin.values[i - 1] + coin.tier_scaling * (coin.values[i] - math.floor(coin.values[i]))
            end
            coin.values[i] = math.floor(coin.values[i])
        end
    end
    
    -- Handle carrying to higher tiers
    for i = 1, coin.max_coin_tier - 1 do
        if coin.values[i] >= tier_scaling then
            local carry = math.floor(coin.values[i] / tier_scaling)
            coin.values[i] = coin.values[i] % tier_scaling
            coin.values[i + 1] = coin.values[i + 1] + carry
        end
    end
    
    -- Handle borrowing from higher tiers for negative values
    for i = 1, coin.max_coin_tier - 1 do
        while coin.values[i] < 0 do
            if coin.values[i + 1] > 0 then
                coin.values[i + 1] = coin.values[i + 1] - 1
                coin.values[i] = coin.values[i] + tier_scaling
            else
                -- Cannot borrow, so zero out all tiers
                for j = 1, coin.max_coin_tier do
                    coin.values[j] = 0
                end
                return coin
            end
        end
    end
    
    -- If highest tier is negative, zero out everything
    if coin.values[coin.max_coin_tier] < 0 then
        for i = 1, coin.max_coin_tier do
            coin.values[i] = 0
        end
        lib.log_error("encountered negative coin value after normalization: " .. serpent.line(coin.values))
    end
    
    return coin
end

-- Add two coin values
function coin_tiers.add(coin1, coin2)
    -- Ensure configurations match
    if coin1.tier_scaling ~= coin2.tier_scaling or coin1.max_coin_tier ~= coin2.max_coin_tier then
        lib.log_error("Cannot add coins with different configurations")
    end
    
    local result = coin_tiers.new()
    
    for i = 1, coin1.max_coin_tier do
        result.values[i] = coin1.values[i] + coin2.values[i]
    end
    
    return coin_tiers.normalize(result)
end

-- Subtract coin2 from coin1
function coin_tiers.subtract(coin1, coin2)
    -- Ensure configurations match
    if coin1.tier_scaling ~= coin2.tier_scaling or coin1.max_coin_tier ~= coin2.max_coin_tier then
        lib.log_error("Cannot subtract coins with different configurations")
    end
    
    local result = coin_tiers.new()
    
    for i = 1, coin1.max_coin_tier do
        result.values[i] = coin1.values[i] - coin2.values[i]
    end
    
    return coin_tiers.normalize(result)
end

-- Multiply a coin value by a number
function coin_tiers.multiply(coin, factor)
    local result = coin_tiers.new()
    
    for i = 1, coin.max_coin_tier do
        result.values[i] = coin.values[i] * factor
    end
    
    return coin_tiers.normalize(result)
end

-- Divide a coin value by a number
function coin_tiers.divide(coin, divisor)
    if divisor == 0 then
        lib.log_error("Cannot divide by zero")
    end
    
    local result = coin_tiers.new()
    local remainder = 0
    
    -- Process from highest tier to lowest
    for i = coin.max_coin_tier, 1, -1 do
        -- Add remainder from higher tier
        local current = coin.values[i] + remainder
        
        -- Divide current tier value
        result.values[i] = math.floor(current / divisor)
        
        -- Calculate remainder and convert to lower tier units
        remainder = (current % divisor) * coin.tier_scaling
    end

    if coin_tiers.is_zero(result) then
        result.values[1] = 1
    end
    
    return result
end

-- Compare two coin values (returns -1 if coin1 < coin2, 0 if equal, 1 if coin1 > coin2)
function coin_tiers.compare(coin1, coin2)
    -- Ensure configurations match
    -- This could technically be handled correctly by converting the tier scaling, but it is not needed for Hextorio.
    -- It can be handled easily by comparing base values (coin_tiers.to_base_value() comparison),
    -- but that involves multiplcation, whereas this method does not.
    if coin1.tier_scaling ~= coin2.tier_scaling or coin1.max_coin_tier ~= coin2.max_coin_tier then
        lib.log_error("Cannot compare coins with different configurations")
    end
    
    for i = coin1.max_coin_tier, 1, -1 do
        if coin1.values[i] < coin2.values[i] then
            return -1
        elseif coin1.values[i] > coin2.values[i] then
            return 1
        end
    end
    
    return 0  -- They are equal
end

function coin_tiers.gt(coin1, coin2)
    return coin_tiers.compare(coin1, coin2) == 1
end

function coin_tiers.lt(coin1, coin2)
    return coin_tiers.compare(coin1, coin2) == -1
end

function coin_tiers.ge(coin1, coin2)
    return coin_tiers.compare(coin1, coin2) >= 0
end

function coin_tiers.le(coin1, coin2)
    return coin_tiers.compare(coin1, coin2) <= 0
end

function coin_tiers.is_zero(coin)
    for i = 1, coin.max_coin_tier do
        if coin.values[i] ~= 0 then
            return false
        end
    end
    return true
end

-- Get the total value in terms of the lowest tier
function coin_tiers.to_base_value(coin)
    local total = 0
    local multiplier = 1
    
    for i = 1, coin.max_coin_tier do
        total = total + coin.values[i] * multiplier
        multiplier = multiplier * coin.tier_scaling
    end
    
    return total
end

-- Create a coin from a value in the lowest tier
function coin_tiers.from_base_value(value, tier_scaling, max_coin_tier)
    local coin = coin_tiers.new()
    coin.tier_scaling = tier_scaling or 100000
    coin.max_coin_tier = max_coin_tier or 4

    local remaining = value
    
    for i = 1, coin.max_coin_tier do
        coin.values[i] = remaining % coin.tier_scaling
        remaining = math.floor(remaining / coin.tier_scaling)
    end
    
    -- If there's remaining value beyond the highest tier, add it to the highest tier
    if remaining > 0 then
        coin.values[coin.max_coin_tier] = coin.values[coin.max_coin_tier] + remaining * coin.tier_scaling
    end
    
    return coin
end

-- Normalize the inventory, combining multiple stacks of coins into their next tiers.
function coin_tiers.normalize_inventory(inventory)
    if storage.coin_tiers.is_processing[inventory] then return end
    storage.coin_tiers.is_processing[inventory] = true

    local total_hex_coins = inventory.get_item_count "hex-coin"
    local total_gravity_coins = inventory.get_item_count "gravity-coin"
    local total_meteor_coins = inventory.get_item_count "meteor-coin"
    local total_hexaprism_coins = inventory.get_item_count "hexaprism-coin"

    if total_hex_coins > 0 or total_gravity_coins > 0 or total_meteor_coins > 0 or total_hexaprism_coins > 0 then
        local coin = coin_tiers.new()
        coin.values = {total_hex_coins, total_gravity_coins, total_meteor_coins, total_hexaprism_coins}
        
        local normalized_coin = coin_tiers.normalize(coin)

        if normalized_coin.values[1] > total_hex_coins then
            inventory.insert {name = "hex-coin", count = normalized_coin.values[1] - total_hex_coins}
        elseif normalized_coin.values[1] < total_hex_coins then
            inventory.remove {name = "hex-coin", count = total_hex_coins - normalized_coin.values[1]}
        end

        if normalized_coin.values[2] > total_gravity_coins then
            inventory.insert {name = "gravity-coin", count = normalized_coin.values[2] - total_gravity_coins}
        elseif normalized_coin.values[2] < total_gravity_coins then
            inventory.remove {name = "gravity-coin", count = total_gravity_coins - normalized_coin.values[2]}
        end

        if normalized_coin.values[3] > total_meteor_coins then
            inventory.insert {name = "meteor-coin", count = normalized_coin.values[3] - total_meteor_coins}
        elseif normalized_coin.values[3] < total_meteor_coins then
            inventory.remove {name = "meteor-coin", count = total_meteor_coins - normalized_coin.values[3]}
        end

        if normalized_coin.values[4] > total_hexaprism_coins then
            inventory.insert {name = "hexaprism-coin", count = normalized_coin.values[4] - total_hexaprism_coins}
        elseif normalized_coin.values[4] < total_hexaprism_coins then
            inventory.remove {name = "hexaprism-coin", count = total_hexaprism_coins - normalized_coin.values[4]}
        end
    end

    storage.coin_tiers.is_processing[inventory] = nil
end



return coin_tiers
