local lib = require "api.lib"
local coin_tiers = require "api.coin_tiers"

local hex_grid = {}

-- Get the third coordinate (s) in the cube coordinate system
-- In axial coords, we only store q,r but sometimes need s for calculations
function hex_grid.get_s(hex)
    return -hex.q - hex.r
end

-- Convert from cube coordinates to axial coordinates
function hex_grid.cube_to_axial(cube)
    return {q = cube.q, r = cube.r}
end

-- Convert from axial coordinates to cube coordinates
function hex_grid.axial_to_cube(hex)
    return {q = hex.q, r = hex.r, s = -hex.q - hex.r}
end

-- Round floating point cube coordinates to the nearest hex
function hex_grid.cube_round(cube)
    local q = math.floor(cube.q + 0.5)
    local r = math.floor(cube.r + 0.5)
    local s = math.floor(cube.s + 0.5)

    local q_diff = math.abs(q - cube.q)
    local r_diff = math.abs(r - cube.r)
    local s_diff = math.abs(s - cube.s)

    -- Fix any rounding errors to ensure q + r + s = 0
    if q_diff > r_diff and q_diff > s_diff then
        q = -r - s
    elseif r_diff > s_diff then
        r = -q - s
    else
        s = -q - r
    end

    return {q = q, r = r, s = s}
end

-- Round floating point axial coordinates to the nearest hex
function hex_grid.round(hex)
    local cube = hex_grid.axial_to_cube(hex)
    local rounded_cube = hex_grid.cube_round(cube)
    return hex_grid.cube_to_axial(rounded_cube)
end

-- Convert rectangular coordinates to axial coordinates
-- rect_pos: {x, y} rectangular coordinates
-- hex_grid_scale: size of hexes (distance from center to corner)
-- hex_grid_rotation: rotation of the grid in radians
function hex_grid.get_hex_containing(rect_pos, hex_grid_scale, hex_grid_rotation)
    -- Default values
    hex_grid_scale = hex_grid_scale or 1
    hex_grid_rotation = hex_grid_rotation or 0
    
    -- Apply rotation if needed
    local x, y = rect_pos.x, rect_pos.y
    if hex_grid_rotation ~= 0 then
        local cos_r = math.cos(-hex_grid_rotation)
        local sin_r = math.sin(-hex_grid_rotation)
        x, y = x * cos_r - y * sin_r, x * sin_r + y * cos_r
    end
    
    -- Convert to cube coordinates using the inverse of the hex to pixel transformation
    local size = hex_grid_scale
    local q = (storage.constants.ROOT_THREE_OVER_THREE * x - 1/3 * y) / size
    local r = (2/3 * y) / size

    -- Create hith floating point coordinates
    local hex = {q = q, r = r}
    -- Round to the nearest hex
    return hex_grid.round(hex)
end

-- Convert axial coordinates to rectangular coordinates (center of hex)
-- hex_pos: {q, r} hex coordinates
-- hex_grid_scale: size of hexes (distance from center to corner)
-- hex_grid_rotation: rotation of the grid in radians
function hex_grid.get_hex_center(hex_pos, hex_grid_scale, hex_grid_rotation)
    -- Default values
    hex_grid_scale = hex_grid_scale or 1
    hex_grid_rotation = hex_grid_rotation or 0
    
    -- Convert hex to pixel
    local size = hex_grid_scale
    local x = size * (storage.constants.ROOT_THREE * hex_pos.q + storage.constants.ROOT_THREE_OVER_TWO * hex_pos.r)
    local y = size * (1.5 * hex_pos.r)
    
    -- Apply rotation if needed
    if hex_grid_rotation ~= 0 then
        local cos_r = math.cos(hex_grid_rotation)
        local sin_r = math.sin(hex_grid_rotation)
        x, y = x * cos_r - y * sin_r, x * sin_r + y * cos_r
    end
    
    return {x = x, y = y}
end

-- Get the six hexes adjacent to the given hex
function hex_grid.get_adjacent_hexes(hex_pos)
    local neighbors = {}
    for i, dir in pairs(storage.hex_grid.directions) do
        table.insert(neighbors, {
            q = hex_pos.q + dir.q,
            r = hex_pos.r + dir.r
        })
    end
    return neighbors
end

-- Calculate the distance between two hexes
function hex_grid.distance(a, b)
    -- In axial coordinates, the distance is calculated as:
    return (math.abs(a.q - b.q) + 
            math.abs(a.q + a.r - b.q - b.r) +
            math.abs(a.r - b.r)) / 2
end

-- Get a line of hexes from a to b
function hex_grid.line(a, b)
    local distance = hex_grid.distance(a, b)
    local results = {}
    
    -- For each point along the line
    for i = 0, distance do
        local t = distance > 0 and i / distance or 0
        
        -- Linear interpolation in cube space
        local ac = hex_grid.axial_to_cube(a)
        local bc = hex_grid.axial_to_cube(b)
        
        -- Interpolate each coordinate separately
        local q = ac.q + (bc.q - ac.q) * t
        local r = ac.r + (bc.r - ac.r) * t
        local s = ac.s + (bc.s - ac.s) * t

        -- Round to nearest hex and convert back to axial
        local cube = hex_grid.cube_round({q = q, r = r, s = s})
        local hex = hex_grid.cube_to_axial(cube)
        
        table.insert(results, hex)
    end
    
    return results
end

-- Get a ring of hexes at a specific radius from the center
function hex_grid.ring(center, radius)
    if radius < 1 then
        return {center}
    end
    
    local results = {}
    
    -- Start at a specific position on the ring
    local hex = {
        q = center.q + storage.hex_grid.directions[4].q * radius,
        r = center.r + storage.hex_grid.directions[4].r * radius
    }
    
    -- Follow the ring by moving in each of the 6 directions
    for i = 1, 6 do
        local dir = i % 6 + 1 -- Wrap around to the next direction (1-based indexing)
        
        -- Move radius hexes in this direction
        for j = 1, radius do
            table.insert(results, {q = hex.q, r = hex.r})
            
            -- Move to the next hex in this direction
            hex.q = hex.q + storage.hex_grid.directions[dir].q
            hex.r = hex.r + storage.hex_grid.directions[dir].r
        end
    end
    
    return results
end

-- Get a spiral of hexes up to a specific radius
function hex_grid.spiral(center, radius)
    local results = {{q = center.q, r = center.r}}
    
    -- Add each ring from 1 to radius
    for r = 1, radius do
        local ring = hex_grid.ring(center, r)
        for _, hex in pairs(ring) do
            table.insert(results, hex)
        end
    end
    
    return results
end

-- Rotate a hex around a center (in 60-degree increments)
function hex_grid.rotate(hex, center, rotation_steps)
    -- Convert to cube coordinates relative to center
    local cube = hex_grid.axial_to_cube(hex)
    local center_cube = hex_grid.axial_to_cube(center)
    
    -- Calculate relative position
    local q = cube.q - center_cube.q
    local r = cube.r - center_cube.r
    local s = cube.s - center_cube.s
    
    -- Normalize rotation_steps to 0-5
    rotation_steps = rotation_steps % 6
    
    -- Rotate in 60-degree increments
    for i = 1, rotation_steps do
        local temp = q
        q = -s
        s = -r
        r = -temp
    end
    
    -- Convert back to axial coordinates and add center offset
    return {
        q = q + center_cube.q,
        r = r + center_cube.r
    }
end

-- Get the corner points of a hex in rectangular coordinates
function hex_grid.get_hex_corners(hex_pos, hex_grid_scale, hex_grid_rotation)
    -- Default values
    hex_grid_scale = hex_grid_scale or 1
    hex_grid_rotation = hex_grid_rotation or 0
    
    -- Get center without rotation
    local center = hex_grid.get_hex_center(hex_pos, hex_grid_scale, 0)
    local corners = {}
    
    -- Calculate the six corners
    for i = 1, 6 do
        local angle = (i - 1) * math.pi / 3 + math.pi / 6 -- +30 degrees offset for flat-top orientation
        local x = center.x + hex_grid_scale * math.cos(angle)
        local y = center.y + hex_grid_scale * math.sin(angle)
        
        -- Apply rotation around the origin if needed
        if hex_grid_rotation ~= 0 then
            local cos_r = math.cos(hex_grid_rotation)
            local sin_r = math.sin(hex_grid_rotation)
            x, y = x * cos_r - y * sin_r, x * sin_r + y * cos_r
        end
        
        table.insert(corners, {x = x, y = y})
    end
    
    return corners
end

-- Check if two hexes are equal
function hex_grid.equals(a, b)
    return a.q == b.q and a.r == b.r
end

-- Calculate a hash value for a hex for use in tables
function hex_grid.hash(hex)
    -- Simple hash function for hex coordinates
    return hex.q * 10000 + hex.r
end

-- Add two hexes together
function hex_grid.add(a, b)
    return {q = a.q + b.q, r = a.r + b.r}
end

-- Subtract hex b from hex a
function hex_grid.subtract(a, b)
    return {q = a.q - b.q, r = a.r - b.r}
end

-- Multiply a hex by a scalar
function hex_grid.multiply(hex, k)
    return {q = hex.q * k, r = hex.r * k}
end

-- Get a random hex within a certain radius of a center hex
function hex_grid.random_hex(center, radius)
    -- Get a random number in the range [-radius, radius]
    local q_offset = math.random(-radius, radius)
    local r_min = math.max(-radius, -q_offset - radius)
    local r_max = math.min(radius, -q_offset + radius)
    local r_offset = math.random(r_min, r_max)
    
    return {
        q = center.q + q_offset,
        r = center.r + r_offset
    }
end

-- Check if a hex is within a certain radius of a center hex
function hex_grid.is_within_radius(hex, center, radius)
    return hex_grid.distance(hex, center) <= radius
end

-- Generate a random rectangular coordinate point within a given hex
function hex_grid.random_rect_in_hex(hex_pos, hex_grid_scale, hex_grid_rotation)
    -- Default values
    hex_grid_scale = hex_grid_scale or 1
    hex_grid_rotation = hex_grid_rotation or 0
    
    -- Get the center and corners of the hex
    local center = hex_grid.get_hex_center(hex_pos, hex_grid_scale, hex_grid_rotation)
    local corners = hex_grid.get_hex_corners(hex_pos, hex_grid_scale, hex_grid_rotation)
    
    -- Instead of rejection sampling, use triangular decomposition
    -- Pick one of the 6 triangles formed by the center and two adjacent corners
    local triangle_idx = math.random(1, 6)
    local corner1 = corners[triangle_idx]
    local corner2 = corners[triangle_idx % 6 + 1] -- Wrap around to first corner if needed
    
    -- Generate random barycentric coordinates
    -- For a random point in a triangle, we need 3 random numbers that sum to 1
    local a = math.random()
    local b = math.random() * (1 - a)
    local c = 1 - a - b
    
    -- Calculate the point using barycentric coordinates
    local x = a * center.x + b * corner1.x + c * corner2.x
    local y = a * center.y + b * corner1.y + c * corner2.y
    
    return {x = x, y = y}
end

-- Helper function to calculate the distance from a point to a line segment
local function point_to_line_distance(point, p1, p2)
    -- Vector from p1 to p2
    local v_x = p2.x - p1.x
    local v_y = p2.y - p1.y
    
    -- Vector from p1 to point
    local w_x = point.x - p1.x
    local w_y = point.y - p1.y
    
    -- Project w onto v
    local c1 = v_x * w_x + v_y * w_y
    if c1 <= 0 then
        -- Point is closest to p1
        return math.sqrt(w_x * w_x + w_y * w_y)
    end
    
    -- Length of v squared
    local c2 = v_x * v_x + v_y * v_y
    if c2 <= c1 then
        -- Point is closest to p2
        local dx = point.x - p2.x
        local dy = point.y - p2.y
        return math.sqrt(dx * dx + dy * dy)
    end
    
    -- Point is closest to a point on the line segment
    local t = c1 / c2
    local proj_x = p1.x + t * v_x
    local proj_y = p1.y + t * v_y
    
    local dx = point.x - proj_x
    local dy = point.y - proj_y
    return math.sqrt(dx * dx + dy * dy)
end

-- Helper function to check if a point is inside a rectangle
local function is_point_in_rect(point, rect_top_left, rect_bottom_right)
    return point.x >= rect_top_left.x and point.x <= rect_bottom_right.x and
           point.y >= rect_top_left.y and point.y <= rect_bottom_right.y
end

-- Helper function to check if a point is inside a polygon (hex)
local function is_point_in_polygon(point, polygon)
    local inside = false
    local j = #polygon
    
    for i = 1, #polygon do
        if ((polygon[i].y > point.y) ~= (polygon[j].y > point.y)) and
           (point.x < (polygon[j].x - polygon[i].x) * (point.y - polygon[i].y) / 
            (polygon[j].y - polygon[i].y) + polygon[i].x) then
            inside = not inside
        end
        j = i
    end
    
    return inside
end

-- Helper function to check if two line segments intersect
local function do_segments_intersect(p1, p2, p3, p4)
    -- Calculate the direction vectors
    local d1x, d1y = p2.x - p1.x, p2.y - p1.y
    local d2x, d2y = p4.x - p3.x, p4.y - p3.y
    
    -- Calculate the determinant
    local det = d1x * d2y - d1y * d2x
    
    -- Lines are parallel if det is close to 0
    if math.abs(det) < 1e-10 then
        return false
    end
    
    -- Calculate the parameters of intersection
    local dx, dy = p1.x - p3.x, p1.y - p3.y
    local t1 = (dx * d2y - dy * d2x) / det
    local t2 = (dx * d1y - dy * d1x) / det
    
    -- Check if the intersection is within both line segments
    return t1 >= 0 and t1 <= 1 and t2 >= 0 and t2 <= 1
end

-- Helper function to check if a line segment intersects a rectangle
local function does_segment_intersect_rect(p1, p2, rect_top_left, rect_bottom_right)
    -- Rectangle corners
    local rect_corners = {
        {x = rect_top_left.x, y = rect_top_left.y},           -- Top left
        {x = rect_bottom_right.x, y = rect_top_left.y},       -- Top right
        {x = rect_bottom_right.x, y = rect_bottom_right.y},   -- Bottom right
        {x = rect_top_left.x, y = rect_bottom_right.y}        -- Bottom left
    }
    
    -- Check intersection with each edge of the rectangle
    for i = 1, 4 do
        local next_i = i % 4 + 1
        if do_segments_intersect(p1, p2, rect_corners[i], rect_corners[next_i]) then
            return true
        end
    end
    
    return false
end

-- Helper function to check if a hex overlaps a rectangle
local function does_hex_overlap_rect(hex_pos, hex_grid_scale, hex_grid_rotation, rect_top_left, rect_bottom_right)
    -- Get the corners of the hex
    local corners = hex_grid.get_hex_corners(hex_pos, hex_grid_scale, hex_grid_rotation)
    
    -- 1. Check if any hex corner is inside the rectangle
    for _, corner in pairs(corners) do
        if is_point_in_rect(corner, rect_top_left, rect_bottom_right) then
            return true
        end
    end
    
    -- 2. Check if any rectangle corner is inside the hex
    local rect_corners = {
        {x = rect_top_left.x, y = rect_top_left.y},           -- Top left
        {x = rect_bottom_right.x, y = rect_top_left.y},       -- Top right
        {x = rect_bottom_right.x, y = rect_bottom_right.y},   -- Bottom right
        {x = rect_top_left.x, y = rect_bottom_right.y}        -- Bottom left
    }
    
    for _, corner in pairs(rect_corners) do
        if is_point_in_polygon(corner, corners) then
            return true
        end
    end
    
    -- 3. Check if any hex edge intersects any rectangle edge
    for i = 1, 6 do
        local next_i = i % 6 + 1
        if does_segment_intersect_rect(corners[i], corners[next_i], rect_top_left, rect_bottom_right) then
            return true
        end
    end
    
    return false
end

-- Get all hexes that overlap a rectangular area
function hex_grid.get_overlapping_hexes(rect_top_left, rect_bottom_right, hex_grid_scale, hex_grid_rotation)
    -- Default values
    hex_grid_scale = hex_grid_scale or 1
    hex_grid_rotation = hex_grid_rotation or 0
    
    -- Normalize the rectangle coordinates (ensure top_left is actually top-left)
    local tl = {
        x = math.min(rect_top_left.x, rect_bottom_right.x),
        y = math.min(rect_top_left.y, rect_bottom_right.y)
    }
    local br = {
        x = math.max(rect_top_left.x, rect_bottom_right.x),
        y = math.max(rect_top_left.y, rect_bottom_right.y)
    }
    
    -- Calculate a margin to accommodate rotated hexes that might overlap
    -- This margin should be at least the hex size plus some buffer for rotation
    local margin = hex_grid_scale * 2
    
    -- Get hexes for the four corners of the rectangle (plus margin)
    local tl_hex = hex_grid.get_hex_containing({x = tl.x - margin, y = tl.y - margin}, hex_grid_scale, hex_grid_rotation)
    local tr_hex = hex_grid.get_hex_containing({x = br.x + margin, y = tl.y - margin}, hex_grid_scale, hex_grid_rotation)
    local bl_hex = hex_grid.get_hex_containing({x = tl.x - margin, y = br.y + margin}, hex_grid_scale, hex_grid_rotation)
    local br_hex = hex_grid.get_hex_containing({x = br.x + margin, y = br.y + margin}, hex_grid_scale, hex_grid_rotation)
    
    -- Find the min/max q and r to create a bounding box of hexes
    local min_q = math.min(tl_hex.q, tr_hex.q, bl_hex.q, br_hex.q)
    local max_q = math.max(tl_hex.q, tr_hex.q, bl_hex.q, br_hex.q)
    local min_r = math.min(tl_hex.r, tr_hex.r, bl_hex.r, br_hex.r)
    local max_r = math.max(tl_hex.r, tr_hex.r, bl_hex.r, br_hex.r)
    
    -- Check each hex in the bounding box
    local overlapping_hexes = {}
    for q = min_q, max_q do
        for r = min_r, max_r do
            local hex = {q = q, r = r}
            if does_hex_overlap_rect(hex, hex_grid_scale, hex_grid_rotation, tl, br) then
                table.insert(overlapping_hexes, hex)
            end
        end
    end
    
    return overlapping_hexes
end

-- Get all integer rectangular coordinates within a hex, excluding the border
function hex_grid.get_hex_tile_positions(hex_pos, hex_grid_scale, hex_grid_rotation, stroke_width)
    -- Default values
    hex_grid_scale = hex_grid_scale or 1
    hex_grid_rotation = hex_grid_rotation or 0
    stroke_width = stroke_width or 0
    
    -- Get the corners of the hex
    local corners = hex_grid.get_hex_corners(hex_pos, hex_grid_scale, hex_grid_rotation)
    
    -- Find the bounding rectangle of the hex
    local min_x, min_y = math.huge, math.huge
    local max_x, max_y = -math.huge, -math.huge
    
    for _, corner in pairs(corners) do
        min_x = math.min(min_x, corner.x)
        min_y = math.min(min_y, corner.y)
        max_x = math.max(max_x, corner.x)
        max_y = math.max(max_y, corner.y)
    end
    
    -- Round to integers (expanding slightly to ensure we cover the full hex)
    min_x = math.floor(min_x - 0.5)
    min_y = math.floor(min_y - 0.5)
    max_x = math.ceil(max_x + 0.5)
    max_y = math.ceil(max_y + 0.5)
    
    -- Collect all integer positions within the hex, excluding border
    local positions = {}
    
    for x = min_x, max_x do
        for y = min_y, max_y do
            local point = {x = x, y = y}
            
            -- Check if the point is inside the hex
            if is_point_in_polygon(point, corners) then
                -- If stroke_width is provided, check distance from edges
                local include = true
                
                if stroke_width > 0 then
                    -- Calculate minimum distance from the point to any edge of the hex
                    local min_distance = math.huge
                    
                    for i = 1, 6 do
                        local next_i = i % 6 + 1
                        local p1 = corners[i]
                        local p2 = corners[next_i]
                        
                        -- Calculate distance from point to line segment
                        local distance = point_to_line_distance(point, p1, p2)
                        min_distance = math.min(min_distance, distance)
                    end
                    
                    -- Exclude points that are within stroke_width of the edge
                    include = min_distance >= stroke_width
                end
                
                if include then
                    table.insert(positions, {x = x, y = y})
                end
            end
        end
    end
    
    return positions
end

----------------------------------------
-- Hex grid generation and management --
----------------------------------------

-- Get or create surface storage
function hex_grid.get_surface_hexes(surface_name)
    local surface_hexes = storage.hex_grid.surface_hexes[surface_name]
    if not surface_hexes then
        surface_hexes = {}
        storage.hex_grid.surface_hexes[surface_name] = surface_hexes
    end
    return surface_hexes
end

-- Get a hex by its axial coordinates in a surface's hex grid.  Defaults and sets to an empty table if the hex does not exist.
function hex_grid.get_hex_in_surface_hexes(surface_hexes, hex_pos)
    local Q = surface_hexes[hex_pos.q]
    if not Q then
        Q = {}
        surface_hexes[hex_pos.q] = Q
    end
    local state = Q[hex_pos.r]
    if not state then
        state = {}
        Q[hex_pos.r] = state
    end
    return state
end

-- Get the state of a hex on a specific surface
function hex_grid.get_hex_state(surface_name, hex_pos)
    local surface_hexes = hex_grid.get_surface_hexes(surface_name)
    local state = hex_grid.get_hex_in_surface_hexes(surface_hexes, hex_pos)
    if not state.position then
        state.position = {q = hex_pos.q, r = hex_pos.r} -- copy position just in case
    end
    return state
end

-- Get the state of a hex from a hex core entity
function hex_grid.get_hex_state_from_core(hex_core)
    if not hex_core then return end
    local transformation = hex_grid.get_surface_transformation(hex_core.surface.name)
    local hex_pos = hex_grid.get_hex_containing(hex_core.position, transformation.scale, transformation.rotation)
    local state = hex_grid.get_hex_state(hex_core.surface.name, hex_pos)
    if state.hex_core ~= hex_core then
        game.print("hex core entities do not match")
        game.print(state.hex_core)
        game.print(hex_core)
    end
    return state
end

-- Set the tile type for a specific position
function hex_grid.set_tile(surface_name, position, tile_type)
    game.surfaces[surface_name].set_tiles({{name = tile_type, position = position}})
end

-- Set the tile type for a list of positions
function hex_grid.set_tiles(surface_name, positions, tile_type)
    local tiles = {}
    for _, position in pairs(positions) do
        table.insert(tiles, {name = tile_type, position = position})
    end
    game.surfaces[surface_name].set_tiles(tiles)
end

-- Set all tiles within a hex to tile_type
function hex_grid.set_hex_tiles(surface_name, hex_pos, tile_type, overwrite_water)
    if tile_type == "none" then return end

    if overwrite_water == nil then
        overwrite_water = true
    end

    local transformation = hex_grid.get_surface_transformation(surface_name)
    local positions = hex_grid.get_hex_tile_positions(hex_pos, transformation.scale, transformation.rotation, transformation.stroke_width)

    hex_grid.set_tiles(surface_name, positions, tile_type)
end

-- Generate tiles along the border of a hex
function hex_grid.generate_hex_border(surface_name, hex_pos, hex_grid_scale, hex_grid_rotation, stroke_width)
    -- Default values
    hex_grid_scale = hex_grid_scale or 1
    hex_grid_rotation = hex_grid_rotation or 0
    stroke_width = stroke_width or 3

    stroke_width = stroke_width * 2.5 -- dividing by the 0.8 factor for tile overlap and multiplying by two to adjust for the halving of stroke width
    
    local corners = hex_grid.get_hex_corners(hex_pos, hex_grid_scale, hex_grid_rotation)
    local border_tiles = {}
    
    -- For each edge of the hex, place water tiles along it
    for i = 1, 6 do
        local next_i = i % 6 + 1
        local start = corners[i]
        local finish = corners[next_i]
        
        -- Calculate the number of steps needed (depends on hex size)
        -- Higher number for smoother border, but more intensive
        local steps = math.max(10, math.floor(hex_grid_scale * 2))
        
        -- Calculate direction vector of the edge
        local dir_x = finish.x - start.x
        local dir_y = finish.y - start.y
        
        -- Calculate perpendicular vector (for stroke width)
        local length = math.sqrt(dir_x * dir_x + dir_y * dir_y)
        local perp_x = -dir_y / length
        local perp_y = dir_x / length
        
        -- For each position along the edge
        for step = 0, steps do
            local t = step / steps
            local base_x = start.x + dir_x * t
            local base_y = start.y + dir_y * t
            
            -- Create tiles for the stroke width
            for w = 0, stroke_width - 1 do
                -- Offset in the perpendicular direction
                -- We want the stroke to go inward from the edge
                local offset = (w - (stroke_width - 1) / 2) * 0.8  -- 0.8 factor for tile overlap
                local x = math.floor(base_x + perp_x * offset + 0.5)
                local y = math.floor(base_y + perp_y * offset + 0.5)
                
                table.insert(border_tiles, {x = x, y = y})
            end
        end
    end
    
    -- Set all border tiles to water
    hex_grid.set_tiles(surface_name, border_tiles, "water")
end

function hex_grid.get_surface_transformation(surface_name)
    local transformations = storage.hex_grid.surface_transformations
    if not transformations[surface_name] then
        transformations[surface_name] = {scale = 20, rotation = 0, stroke_width = 5}
    end
    local transformation = transformations[surface_name]
    if not transformation.scale then
        transformation.scale = lib.startup_setting_value "hex-size"
    end
    if not transformation.rotation then
        transformation.rotation = math.random() * math.pi
    end
    if not transformation.stroke_width then
        transformation.stroke_width = lib.startup_setting_value "hex-stroke-width"
    end
    return transformation
end

-- Initialize a hex with default state and generate its border
function hex_grid.initialize_hex(surface_name, hex_pos, hex_grid_scale, hex_grid_rotation, stroke_width)
    local hex = hex_grid.get_hex_state(surface_name, hex_pos)

    -- Skip if this hex has already been generated
    if hex.generated then
        return
    end

    hex.generated = true
    hex_grid.generate_hex_border(surface_name, hex_pos, hex_grid_scale, hex_grid_rotation, stroke_width)
end

-- Check if a hex core can be spawned within a hex
function hex_grid.can_hex_core_spawn(surface_name, hex_pos)
    local state = hex_grid.get_hex_state(surface_name, hex_pos)
    if state.hex_core then
        return false
    end
    if hex_pos.q == 0 and hex_pos.r == 0 then
        return true
    end
    if hex_grid.is_hex_near_claimed_hex(surface_name, hex_pos) then
        return true
    end
    return false
end

-- Check if a hex is near a claimed hex
function hex_grid.is_hex_near_claimed_hex(surface_name, hex_pos)
    local adjacent_hexes = hex_grid.get_adjacent_hexes(hex_pos)
    for _, adj_hex in pairs(adjacent_hexes) do
        local state = hex_grid.get_hex_state(surface_name, adj_hex)
        if state.claimed then
            return true
        end
    end
    return false
end

-- Claim a hex and spawn hex cores in adjacent hexes if possible.
function hex_grid.claim_hex(surface_name, hex_pos)
    local state = hex_grid.get_hex_state(surface_name, hex_pos)
    if state.claimed then return end
    state.claimed = true

    local adjacent_hexes = hex_grid.get_adjacent_hexes(hex_pos)
    local transformation = hex_grid.get_surface_transformation(surface_name)
    for _, adj_hex in pairs(adjacent_hexes) do
        if hex_grid.can_hex_core_spawn(surface_name, adj_hex) then
            hex_grid.spawn_hex_core(surface_name, hex_grid.get_hex_center(adj_hex, transformation.scale, transformation.rotation))
        end
    end

    -- Set tiles
    hex_grid.set_hex_tiles(surface_name, hex_pos, lib.startup_setting_value "claimed-hex-tile")
    log("claimed hex at " .. serpent.line(hex_pos) .. " / " .. serpent.line(hex_grid.get_hex_center(hex_pos, transformation.scale, transformation.rotation)))
end

-- Handle chunk generation event for the hex grid
function hex_grid.on_chunk_generated(surface_name, chunk_pos, hex_grid_scale, hex_grid_rotation, stroke_width)
    -- Default values
    local transformation = hex_grid.get_surface_transformation(surface_name)
    hex_grid_scale = transformation.scale
    hex_grid_rotation = transformation.rotation
    stroke_width = transformation.stroke_width
    
    -- Convert chunk position to rectangle coordinates
    local top_left, bottom_right = lib.chunk_to_rect(chunk_pos)
    
    -- Find all hexes that overlap with this chunk
    local overlapping_hexes = hex_grid.get_overlapping_hexes(
        top_left, bottom_right, hex_grid_scale, hex_grid_rotation
    )
    
    -- Initialize each overlapping hex if not already generated
    for _, hex_pos in pairs(overlapping_hexes) do
        hex_grid.initialize_hex(surface_name, hex_pos, hex_grid_scale, hex_grid_rotation, stroke_width)

        local center = hex_grid.get_hex_center(hex_pos, hex_grid_scale, hex_grid_rotation)
        if lib.is_position_in_rect(center, top_left, bottom_right) and hex_grid.can_hex_core_spawn(surface_name, hex_pos) then
            hex_grid.spawn_hex_core(surface_name, center)
        end
    end
    
    -- Return the hexes that were processed for this chunk
    return overlapping_hexes
end

function hex_grid.spawn_hex_core(surface_name, position)
    local surface = game.surfaces[surface_name]
    if not surface then return end

    local entities = surface.find_entities_filtered {
        area = {{position.x - 2.5, position.y - 2.5}, {position.x + 2.5, position.y + 2.5}},
    }
    for _, e in pairs(entities) do
        if e.type ~= "character" then
            e.destroy()
        end
    end

    local hex_core = surface.create_entity {name = "hex-core", position = position, force = "player"}
    surface.create_entity {name = "hex-core-loader", position = {position.x - 1, position.y + 2}, direction = 0, type = "input", force = "player"}
    surface.create_entity {name = "hex-core-loader", position = {position.x + 1, position.y + 2}, direction = 8, type = "output", force = "player"}

    for _, e in pairs(entities) do
        if e.valid and e.type == "character" then
            lib.unstuck_player(e.player)
        end
    end

    local transformation = hex_grid.get_surface_transformation(surface_name)
    local hex_pos = hex_grid.get_hex_containing(position, transformation.scale, transformation.rotation)
    local state = hex_grid.get_hex_state(surface_name, hex_pos)
    state.hex_core = hex_core
    state.claim_price = coin_tiers.from_base_value(1)

    return hex_core
end



return hex_grid
